from __future__ import annotations

import textwrap
import typing as t
import uuid

import click

from globus_sdk.paging import Paginator

from gcs_cli.helpers import (
    requires_login,
)
from gcs_cli.parsing import (
    collection_filter_arg,
    globus_command,
    page_size_opt,
)
from gcs_cli.services.gcs import get_client


@globus_command("check", short_help="Check collection configuration")
@click.argument("collection-id", type=click.UUID, required=False)
@click.option(
    "--mapped-collection-id",
    default=None,
    type=click.UUID,
    help=(
        "Filter results to Guest Collections on a specific Mapped Collection. "
        "This is the ID of the Mapped Collection"
    ),
)
@click.option(
    "--storage-gateway-id",
    default=None,
    type=click.UUID,
    help=(
        "Filter results to Collections on a specific Storage Gateway. "
        "This is the ID of the Storage Gateway"
    ),
)
@collection_filter_arg
@page_size_opt
@requires_login
def collection_check(
    mapped_collection_id: uuid.UUID | None,
    storage_gateway_id: uuid.UUID | None,
    filters: t.Sequence[str],
    collection_id: uuid.UUID | None,
    page_size: int | None,
) -> None:
    """
    Check one or more collections for consistency.
    """
    client = get_client()

    params: dict[str, str | int] = {}
    if mapped_collection_id:
        params["mapped_collection_id"] = str(mapped_collection_id)
    if storage_gateway_id:
        params["storage_gateway_id"] = str(storage_gateway_id)
    if filters:
        params["filter"] = ",".join(filters)
    if page_size:
        params["page_size"] = page_size

    if params and collection_id:
        click.echo("Either provide collection filters or a collection id", err=True)
        click.get_current_context().exit(1)

    if collection_id:
        res_data = client.check_collection(collection_id).data["data"]
    else:
        res: Paginator = client.paginated.check_collections(**params)
        res_data = [r for r in res.items()]

    errors_found: dict[str, t.Any] = {}
    for datum in res_data:
        error = datum.get("error")
        if error:
            data_type = error["DATA_TYPE"]
            if data_type not in errors_found:
                errors_found[data_type] = []
            errors_found[data_type].append(datum)

    if errors_found:
        for error_type in errors_found:
            for e in errors_found[error_type]:
                click.echo(e["id"])
                click.echo("\n    ".join(textwrap.wrap("    " + e["message"])))
                click.echo()
    else:
        click.echo("No errors detected")
