from __future__ import annotations

import typing as t
import uuid

import click

from gcs_cli.helpers import requires_login
from gcs_cli.parsing import (
    activity_notifications_opt,
    globus_command,
)
from gcs_cli.prettyprint import FORMAT_TEXT_RECORD, formatted_print
from gcs_cli.sdk_client.data import CollectionDocument
from gcs_cli.services.gcs import get_client


@globus_command("create-guest", short_help="Create a new Guest Collection", hidden=True)
@click.argument("MAPPED_COLLECTION_ID")
@click.argument("CREDENTIAL_ID")
@click.argument("BASE_PATH")
@click.argument("DISPLAY_NAME")
@click.option(
    "--public/--private",
    "public",
    help="Set the collection to be private or public (defaults to public)",
    default=True,
)
@click.option(
    "--skip-auto-delete",
    is_flag=True,
    default=None,
    help=(
        "Option indicating whether the guest collection is subject to "
        "automatic deletion if --auto-delete-timeout is set on its mapped "
        "collection. Only settable on guest collections. Defaults to false"
    ),
)
@activity_notifications_opt(True)
@requires_login
def collection_create_guest(
    mapped_collection_id: uuid.UUID,
    credential_id: uuid.UUID,
    base_path: str,
    display_name: str,
    public: bool,
    activity_notifications: list[str] | None,
    skip_auto_delete: bool | None = None,
) -> None:
    """
    Create a new Guest Collection, rooted on some given path within an
    existing Mapped Collection.

    The MAPPED_COLLECTION_ID is the Mapped Collection which will host this
    Guest Collection.

    [WARNING]: For Internal Use Only
    This command may change in future versions without notice.
    """
    # By making use of kwargs instead of directly adding values to the
    # CollectionDocument() call, we minimize the value of DATA_TYPE which in turn
    # keeps the CLI more compatible with older GCS installations.
    kwargs: dict[str, t.Any] = {}
    # 5.4.82 / collection#1.13.0
    if skip_auto_delete is True:
        kwargs["skip_auto_delete"] = True

    collection_doc = CollectionDocument(
        collection_type="guest",
        mapped_collection_id=mapped_collection_id,
        user_credential_id=credential_id,
        collection_base_path=base_path,
        display_name=display_name,
        public=public,
        activity_notifications=activity_notifications,
        **kwargs,  # type: ignore
    )
    client = get_client()
    res = client.create_collection(collection_doc)

    # output
    formatted_print(
        res, fields=[("Collection ID", "id")], text_format=FORMAT_TEXT_RECORD
    )
