from __future__ import annotations

import typing as t
import uuid

import click
from typing_extensions import Literal

from globus_sdk import GlobusAPIError

from gcs_cli.commands.collection._associated_flow_policy import (
    make_associated_flow_policy,
)
from gcs_cli.helpers import requires_login
from gcs_cli.parsing import (
    collection_create_and_update_params,
    globus_command,
    mapped_collection_create_and_update_params,
)
from gcs_cli.parsing.types import (
    EXPLICIT_NULL,
    _ExplicitNullClass,
)
from gcs_cli.prettyprint import FORMAT_TEXT_RECORD, formatted_print
from gcs_cli.sdk_client.data import (
    CollectionDocument,
    GoogleCloudStorageCollectionPolicies,
    POSIXCollectionPolicies,
    POSIXStagingCollectionPolicies,
)
from gcs_cli.services.gcs import get_client


@globus_command("create", short_help="Create a new Mapped Collection")
@click.argument("STORAGE_GATEWAY_ID")
@click.argument("BASE_PATH")
@collection_create_and_update_params(create=True)
@mapped_collection_create_and_update_params
@click.option(
    "--identity-id",
    help=(
        "Globus Auth identity to who acts as the owner of this Collection. This "
        "identity is an administrator on the Endpoint. This ID must be in your "
        "current identity set"
    ),
)
@click.option(
    "--delete-protected",
    is_flag=True,
    help=(
        "If --delete-protected, then this collection can not be deleted "
        "until --no-delete-protected is used with the collection update command. "
        "This option is enabled by default as of GCS v5.4.69."
    ),
)
@requires_login
def collection_create_mapped(
    storage_gateway_id: uuid.UUID,
    base_path: str,
    identity_id: str | None,
    delete_protected: bool,
    google_project_id: str | None,
    acl_expiration_mins: int | None = None,
    posix_sharing_group_allow: t.Iterable[str] | None = None,
    posix_sharing_group_deny: t.Iterable[str] | None = None,
    posix_staging_sharing_group_allow: t.Iterable[str] | None = None,
    posix_staging_sharing_group_deny: t.Iterable[str] | None = None,
    sharing_restrict_paths: dict[str, t.Iterable[str]] | None = None,
    sharing_users_allow: t.Iterable[str] | None = None,
    sharing_users_deny: t.Iterable[str] | None = None,
    user_message: str | None = None,
    user_message_link: str | None = None,
    verify: Literal["force"] | Literal["disable"] | Literal["default"] | None = None,
    guest_auth_policy_id: uuid.UUID | None = None,
    restrict_transfers_to_high_assurance: str | None = None,
    auto_delete_timeout: int | None = None,
    flow_transfer_source: uuid.UUID | _ExplicitNullClass | None = None,
    flow_transfer_destination: uuid.UUID | _ExplicitNullClass | None = None,
    **kwargs: t.Any,
) -> None:
    """
    Create a new Mapped Collection, rooted on some given path within an
    existing Storage Gateway.

    The STORAGE_GATEWAY_ID is the Storage Gateway which will host the Collection.
    """
    # 5.4.82 / collection#1.13.0
    if auto_delete_timeout is not None:
        kwargs["auto_delete_timeout"] = auto_delete_timeout

    if delete_protected:
        kwargs["delete_protected"] = True
    if identity_id is not None:
        kwargs["identity_id"] = identity_id
    if google_project_id is not None:
        kwargs["policies"] = GoogleCloudStorageCollectionPolicies(
            project=google_project_id
        )

    if sharing_users_allow is not None:
        kwargs["sharing_users_allow"] = sharing_users_allow

    if sharing_users_deny is not None:
        kwargs["sharing_users_deny"] = sharing_users_deny

    if posix_sharing_group_allow is not None or posix_sharing_group_deny is not None:
        # Added in 5.4.8 for POSIX connector
        policies = {}
        if posix_sharing_group_allow is not None:
            policies["sharing_groups_allow"] = posix_sharing_group_allow

        if posix_sharing_group_deny is not None:
            policies["sharing_groups_deny"] = posix_sharing_group_deny
        kwargs["policies"] = POSIXCollectionPolicies(**policies)

    if (
        posix_staging_sharing_group_allow is not None
        or posix_staging_sharing_group_deny is not None
    ):
        # Added in 5.4.10 for POSIX staging connector
        policies = {}
        if posix_staging_sharing_group_allow is not None:
            policies["sharing_groups_allow"] = posix_staging_sharing_group_allow

        if posix_staging_sharing_group_deny is not None:
            policies["sharing_groups_deny"] = posix_staging_sharing_group_deny
        kwargs["policies"] = POSIXStagingCollectionPolicies(**policies)

    if sharing_restrict_paths is not None:
        if sharing_restrict_paths == EXPLICIT_NULL:
            kwargs["sharing_restrict_paths"] = None
        else:
            kwargs["sharing_restrict_paths"] = sharing_restrict_paths

    if acl_expiration_mins is not None:
        if acl_expiration_mins == EXPLICIT_NULL:
            kwargs["acl_expiration_mins"] = None
        else:
            kwargs["acl_expiration_mins"] = acl_expiration_mins

    if user_message is not None:
        if user_message == EXPLICIT_NULL:
            kwargs["user_message"] = None
        else:
            kwargs["user_message"] = user_message

    if user_message_link is not None:
        if user_message_link == EXPLICIT_NULL:
            kwargs["user_message_link"] = None
        else:
            kwargs["user_message_link"] = user_message_link

    if guest_auth_policy_id is not None:
        if guest_auth_policy_id == EXPLICIT_NULL:
            kwargs["guest_auth_policy_id"] = None
        else:
            kwargs["guest_auth_policy_id"] = str(guest_auth_policy_id)

    if restrict_transfers_to_high_assurance is not None:
        if restrict_transfers_to_high_assurance == EXPLICIT_NULL:
            kwargs["restrict_transfers_to_high_assurance"] = None
        else:
            kwargs[
                "restrict_transfers_to_high_assurance"
            ] = restrict_transfers_to_high_assurance

    associated_flow_policy: dict[str, t.Any] | None = make_associated_flow_policy(
        flow_transfer_source,
        flow_transfer_destination,
    )
    if associated_flow_policy:
        kwargs["associated_flow_policy"] = associated_flow_policy

    # These are added in GCS 5.4.4, so if neither --enable-https or
    # --disable-https are set, then we'll not try to set the value for
    # backward compatibility
    if kwargs.get("enable_https") and kwargs.get("disable_https"):
        raise click.exceptions.UsageError(
            "Only one of --enable-https and --disable-https may be used"
        )
    if kwargs.get("enable_https") is False:
        kwargs.pop("enable_https")

    if kwargs.get("disable_https"):
        kwargs["enable_https"] = False

    kwargs.pop("disable_https", None)

    if verify is not None:
        if verify.lower() == "force":
            kwargs["force_verify"] = True
            kwargs["disable_verify"] = False
        elif verify.lower() == "disable":
            kwargs["force_verify"] = False
            kwargs["disable_verify"] = True
        else:
            kwargs["force_verify"] = False
            kwargs["disable_verify"] = False

    collection_doc = CollectionDocument(
        storage_gateway_id=storage_gateway_id,
        collection_type="mapped",
        collection_base_path=base_path,
        **kwargs,
    )
    client = get_client()

    try:
        res = client.create_collection(collection_doc)
    except GlobusAPIError as e:
        # Check if the caller used a connector-specific option for the wrong storage
        # gateway type.
        if e.http_status == 422 and "policies" in kwargs:
            raise click.ClickException(
                "This request used an option not supported by the storage gateway type. "
                "Review the options and try again.\n"
                f'The service responded with "{e.message}"'
            )
        raise e

    # output
    formatted_print(
        res, fields=[("Collection ID", "id")], text_format=FORMAT_TEXT_RECORD
    )
