from __future__ import annotations

import glob
import json
import os
import shlex
import shutil
import socket
import subprocess
import textwrap
import uuid
from functools import cache
from typing import Callable, Iterable, List, Union

import click
import distro
import requests

from gcs_cli.helpers import is_logged_in
from gcs_cli.parsing import globus_command
from gcs_cli.sdk_client.host_header_ssl_adapter import HostHeaderSSLAdapter


def _run_diagnostic_step(
    cmd: Union[List[str], str, Callable], force_try_all: bool
) -> None:
    stagename = (
        str(cmd)
        if not callable(cmd)
        else getattr(cmd, "stagename", "func:{}()".format(cmd.__name__))
    )
    click.secho("== diagnostic: {} ==".format(stagename), fg="yellow", bold=True)

    try:
        if callable(cmd):
            result = cmd()
            if isinstance(result, str):
                click.echo(result)
            return

        if isinstance(cmd, str):
            cmd = shlex.split(cmd)
        arg0 = cmd[0]
        if force_try_all or shutil.which(arg0):
            try:
                subprocess.run(cmd, timeout=30)
            except subprocess.TimeoutExpired:
                click.secho("command timed out", fg="red", bold=True)
        else:
            click.secho("{} was not found in the PATH".format(arg0), fg="red")
    # at the very least, subprocess execution can raise an OSError
    # when attempting to execute a non-executable file or when the command name
    # can't be resolved
    # any errors here should be caught to avoid crashing the whole diagnostic run
    except Exception as e:
        click.secho("command failed: {}".format(e), fg="red", bold=True)


def _linux_distro() -> str | None:
    platforms = {
        "redhat": [
            "almalinux",
            "centos",
            "fedora",
            "oracle",
            "red hat",
            "redhat",
            "rocky",
            "scientific",
            "springdale",
        ],
        "debian": ["debian", "ubuntu"],
        "suse": ["suse"],
    }
    forced = os.environ.get("GCS_FORCE_PLATFORM", "").lower()
    if forced in platforms.keys():
        return forced
    else:
        name = distro.id().lower()
        for p in platforms:
            if [x for x in platforms[p] if name.startswith(x)]:
                return p
    return None


def _is_debian() -> bool:
    """return true if the platform is a debian-like system
    false if it is not or if detection fails in any way"""
    return _linux_distro() == "debian"


def _is_redhat() -> bool:
    """return true if the platform is a redhat-derived system
    false if it is not or if detection fails in any way"""
    return _linux_distro() == "redhat"


def cat(path: str, wildcard: bool = False) -> Callable[[], None]:
    def show_step() -> None:
        if not wildcard:
            files = [path]
        else:
            files = glob.glob(path)

        for filename in files:
            click.echo("cat " + filename)
            if os.path.exists(filename):
                with open(filename) as f:
                    click.echo(textwrap.indent(f.read(), "  "))
            else:
                click.echo("no file named {}".format(filename))

    show_step.stagename = "func:cat({})".format(path)  # type: ignore

    return show_step


def systemctl_status(service: str) -> str:
    return "systemctl status {} --no-pager --lines 10".format(service)


def get_fqdn() -> str:
    return socket.getfqdn()


def user_is_logged_in() -> str:
    return str(is_logged_in())


def running_as_root() -> bool:
    return os.geteuid() == 0


def get_service_names(force_try_all: bool) -> Iterable[str]:
    services = [
        "gcs_manager",
        "gcs_manager.socket",
        "gcs_manager_assistant",
        "globus-gridftp-server",
    ]
    # add apache2 or httpd depending on platform in use
    if force_try_all:
        return ["apache2", "httpd"] + services
    if _is_debian():
        return ["apache2"] + services
    return ["httpd"] + services


def get_info_uri() -> str:
    try:
        with open("/var/lib/globus-connect-server/info.json", "r") as f:
            info = json.load(f)
            return "https://{}/api/info".format(info["domain_name"])
    except Exception as e:
        return click.style(
            "Parsing info.json failed: {}".format(e), fg="red", bold=True
        )


def resolve_nodes() -> str:
    try:
        with open("/var/lib/globus-connect-server/info.json", "r") as f:
            info = json.load(f)
    except Exception as e:
        return click.style(
            "Parsing info.json failed: {}".format(e), fg="red", bold=True
        )
    domain_name = info["domain_name"]

    res = []
    for addr in socket.getaddrinfo(domain_name, 443, proto=socket.IPPROTO_TCP):
        res.append(addr[4][0])

    if len(res) == 0:
        res.append(
            click.style(
                "Domain {} does not resolve to any addresses".format(domain_name),
                fg="red",
                bold=True,
            )
        )
    return "Domain {} resolves to {}".format(domain_name, ", ".join(res))


def contact_info_uri() -> str:
    try:
        with open("/var/lib/globus-connect-server/info.json", "r") as f:
            info = json.load(f)
    except Exception as e:
        return click.style(
            "Parsing info.json failed: {}".format(e), fg="red", bold=True
        )

    domain_name = info["domain_name"]
    client = requests.sessions.Session()
    client.mount("https://", HostHeaderSSLAdapter())

    res = []
    for addr in socket.getaddrinfo(domain_name, 443, proto=socket.IPPROTO_TCP):
        ip_address = addr[4][0]
        try:
            if addr[0] == socket.AF_INET6:
                urlsafe_ip = "[{}]".format(ip_address)
            else:
                urlsafe_ip = ip_address
            info_data = client.get(
                "https://{}/api/info".format(urlsafe_ip),
                headers={"Host": domain_name},
                timeout=1,
            )
            info_data.raise_for_status()
            for datum in info_data.json()["data"]:
                if datum["DATA_TYPE"].startswith("info#"):
                    res.append(
                        click.style(
                            "{}: {}".format(ip_address, json.dumps(datum, indent=2))
                        )
                    )
        except Exception as e:
            res.append(
                "Error accessing GCS Manager at {}: {}".format(
                    ip_address, click.style(str(e), fg="red", bold=True)
                )
            )
    return "\n".join(res)


def contact_gcs_data_service() -> str:
    url = "https://config-data.gcs.globus.org/graphql"
    body = {"query": "{ Endpoint }"}
    client = requests.sessions.Session()
    res = client.post(url, json=body)
    if res.status_code != 401:
        return f"Expected 401, got {res.status_code}"
    res_json = res.json()

    errors = res_json.get("errors", [])
    if len(errors) != 1 or errors[0].get("errorType") != "UnauthorizedException":
        return f"Expected UnauthorizedException, got {errors}"

    return "ok"


fetched_storage_gateways: set[str] = set()


@cache
def gcs_command() -> bool:
    return shutil.which("globus-connect-server") is not None


def get_storage_gateway_info(storage_gateway_id: str, force_try_all: bool) -> str:
    if storage_gateway_id not in fetched_storage_gateways:
        cmd = [
            "globus-connect-server",
            "storage-gateway",
            "show",
            storage_gateway_id,
            "--include-private-policies",
            "-F",
            "json",
        ]
        output = []

        if force_try_all or gcs_command():
            try:
                proc = subprocess.run(
                    cmd, timeout=30, stdout=subprocess.PIPE, text=True
                )
                fetched_storage_gateways.add(storage_gateway_id)
                storage_gateway_json = json.loads(proc.stdout)
                output.append(f"Storage gateway info for {storage_gateway_id}")
                output.append(json.dumps(storage_gateway_json, indent=2))
            except subprocess.TimeoutExpired:
                click.secho("command timed out", fg="red", bold=True)

        return "\n".join(output)
    else:
        return ""


fetched_collections: set[str] = set()


class GetCollectionInfo:
    def __init__(self, collection_id: str | None, force_try_all: bool) -> None:
        self.collection_id = collection_id
        self.force_try_all = force_try_all

    @property
    def __name__(self) -> str:
        return f"GetCollectionInfo(collection_id={self.collection_id}, force_try_all={self.force_try_all})"

    def __call__(self) -> str:
        collections: list[str] = []

        if self.collection_id is None:
            cmd = [
                "globus-connect-server",
                "collection",
                "list",
                "-F",
                "list",
            ]
            try:
                proc = subprocess.run(
                    cmd, timeout=30, stdout=subprocess.PIPE, text=True
                )
                if proc.returncode != 0:
                    return ""

                collections.extend(proc.stdout.split())
            except subprocess.TimeoutExpired:
                click.secho("command timed out", fg="red", bold=True)
        elif self.collection_id not in fetched_collections:
            collections.append(self.collection_id)

        output: list[str] = []
        for collection_id in collections:
            if collection_id not in fetched_collections:
                cmd = [
                    "globus-connect-server",
                    "collection",
                    "show",
                    collection_id,
                    "--include-private-policies",
                    "-F",
                    "json",
                ]

                if self.force_try_all or gcs_command():
                    try:
                        proc = subprocess.run(
                            cmd, timeout=30, stdout=subprocess.PIPE, text=True
                        )
                        if proc.returncode != 0:
                            return ""

                        fetched_collections.add(collection_id)
                        collection_json = json.loads(proc.stdout)
                        storage_gateway_id: str = collection_json["storage_gateway_id"]
                        output.append(f"Collection info for {collection_id}")
                        output.append(json.dumps(collection_json, indent=2))

                        if collection_json["collection_type"] == "guest":
                            mapped_collection_id = collection_json[
                                "mapped_collection_id"
                            ]
                            output.append(
                                GetCollectionInfo(
                                    mapped_collection_id, self.force_try_all
                                )()
                            )
                        else:
                            storage_gateway = get_storage_gateway_info(
                                storage_gateway_id, self.force_try_all
                            )
                            if storage_gateway:
                                output.append(storage_gateway)
                    except subprocess.TimeoutExpired:
                        click.secho("command timed out", fg="red", bold=True)

        return "\n".join(output)


@globus_command(
    "self-diagnostic", disable_options=["map_http_status", "format", "explicit_host"]
)
@click.option(
    "--force-try-all",
    is_flag=True,
    help="Don't skip any steps that are expected to fail",
)
@click.option(
    "--collection",
    multiple=True,
    metavar="UUID|all",
    help="Collection to diagnose",
)
def self_diagnostic_command(force_try_all: bool, collection: tuple[str]) -> None:
    """Run several diagnostic utilities which can be used to provide feedback to the
    Globus Team.

    This tool may produce a large amount of output.
    """
    any_user_diagnostics: List[Union[str, Callable]] = [
        "uname -a",
        "sestatus",
        "ip addr",
        get_fqdn,
        cat("/etc/os-release"),
        cat("/var/lib/globus-connect-server/info.json"),
        user_is_logged_in,
        "globus-gridftp-server --version",
        get_info_uri,
        resolve_nodes,
        contact_info_uri,
        contact_gcs_data_service,
        "getent passwd gcsweb",
        "getent passwd globusoidc",
    ]
    any_user_diagnostics.extend(
        [systemctl_status(x) for x in get_service_names(force_try_all)]
    )
    root_user_diagnostics: List[Union[str, Callable]] = [
        cat("/etc/gridftp.conf"),
        cat("/etc/gridftp.d/*", wildcard=True),
        "ls -l /var/lib/globusoidc",
        "ls -l /var/lib/globus-connect-server/gcs-manager",
    ]
    if _is_redhat():
        any_user_diagnostics.append("rpm -q selinux-policy")
        root_user_diagnostics.append("semodule -lfull")

    if collection == ("all",):
        any_user_diagnostics.append(GetCollectionInfo(None, force_try_all))
    else:
        for col in collection:
            try:
                _ = uuid.UUID(col)
                any_user_diagnostics.append(GetCollectionInfo(col, force_try_all))
            except ValueError:
                raise click.UsageError(f"Invalid collection id {col}")

    for cmd in any_user_diagnostics:
        _run_diagnostic_step(cmd, force_try_all)

    is_root = running_as_root()
    click.secho("= check running_as_root:{} =".format(is_root), fg="yellow", bold=True)
    if is_root or force_try_all:
        for cmd in root_user_diagnostics:
            _run_diagnostic_step(cmd, force_try_all)
