import typing

import click

import globus_sdk

import gcs_cli.helpers as cli_helpers
import gcs_cli.parsing as cli_parsing
import gcs_cli.sdk_client.data as sdkdata
import gcs_cli.services.gcs as cli_servives


@cli_parsing.globus_command(
    "s3-key-add",
    disable_options=["format"],
    short_help="Add IAM keys to an S3 User Credential",
)
@click.option(
    "--s3-secret-key",
    help="S3 secret key. If not provided, this command will prompt for it",
    prompt=True,
    envvar="S3_SECRET_KEY",
)
@click.option(
    "--prefix",
    default=None,
    multiple=True,
    help="The path prefix of all S3 bucket/object path to use with this key. "
    "This command is additive and can be run multiple times.",
)
@cli_parsing.user_credential_id_arg
@click.argument("s3_access_key_id", metavar="S3_ACCESS_KEY_ID", type=str)
@cli_helpers.requires_login
def s3_keys_add(
    prefix: typing.List[str],
    s3_access_key_id: str,
    s3_secret_key: str,
    user_credential_id: str,
) -> None:
    client = cli_servives.get_client()

    key_entry = sdkdata.S3MultiKeyEntry(
        s3_key_id=s3_access_key_id, s3_secret_key=s3_secret_key, path_prefixes=prefix
    )
    policy = sdkdata.S3UserCredentialPolicies(
        DATA_TYPE="s3_user_credential_policies#1.2.0", s3_multi_keys=[key_entry]
    )
    user_cred = sdkdata.UserCredentialDocument(
        DATA_TYPE="user_credential#1.0.0", policies=policy
    )

    try:
        res = client.update_user_credential(user_credential_id, user_cred)
        click.echo(f"Updated credential {res.data['data'][0]['id']}")
    except globus_sdk.GCSAPIError as e:
        if e.http_status == 404:
            raise click.ClickException(
                f"The credential with ID {user_credential_id} was not found"
            )
        else:
            raise click.ClickException(
                f"Failed to update the key {s3_access_key_id}: {getattr(e, 'message', e.detail)}"
            )

    return None
