from __future__ import annotations

import typing as t

import globus.manager.api.schemas.connectors.s3 as schema_s3
from globus.manager.constants import ConfigurationObjectTag

from .decorators import checks_policy_property
from .user_credential import UserCredential


def _validate_s3_credential_in_policy(policy, patch=False):
    if "s3_multi_keys" not in policy:
        return

    if not patch:
        needed_properties = ["s3_key_id", "s3_secret_key", "path_prefixes"]
    else:
        needed_properties = ["s3_key_id"]

    for s3_cred in policy["s3_multi_keys"]:
        for p in needed_properties:
            if p not in s3_cred:
                raise ValueError(
                    f"The property {p} requires a value in every key entry"
                )
            if p == "path_prefixes":
                if (
                    len(s3_cred["path_prefixes"]) > schema_s3.MAX_S3_CREDENTIAL_PREFIXES
                    or len(s3_cred["path_prefixes"]) < 1
                ):
                    raise ValueError(
                        f"At least 1 and at most {schema_s3.MAX_S3_CREDENTIAL_PREFIXES} path_prefixes can be"
                        f" associated with a credential"
                    )


class S3UserCredential(UserCredential):
    POLICY_DATA_TYPE = "s3_user_credential_policies#1.2.0"
    CONNECTOR_POLICIES = {
        "s3_user_credential_policies#1.0.0": [
            "s3_key_id",
            "s3_secret_key",
        ]
    }
    CONNECTOR_POLICIES["s3_user_credential_policies#1.1.0"] = CONNECTOR_POLICIES[
        "s3_user_credential_policies#1.0.0"
    ] + ["s3_requester_pays"]
    CONNECTOR_POLICIES["s3_user_credential_policies#1.2.0"] = CONNECTOR_POLICIES[
        "s3_user_credential_policies#1.1.0"
    ] + ["s3_multi_keys"]

    POLICY_DATA_TYPES = list(CONNECTOR_POLICIES.keys())
    CONNECTOR_ID = "7643e831-5f6c-4b47-a07f-8ee90f401d23"

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.USER_CREDENTIAL.name + CONNECTOR_ID
        )
    }

    anonymous_display_name = "Anonymous S3"
    unconfigured_display_name = "Unprovisioned S3 Credential"

    @property
    def connector_properties(self) -> list[str]:
        return self.CONNECTOR_POLICIES[self.POLICY_DATA_TYPE]

    def __init__(self, appsync_data=None, doc=None, storage_gateway=None):
        if doc:
            invalid = False
            provisioned = False
            policies = doc.get("policies", {})
            s3_key_id = policies.get("s3_key_id")
            s3_secret_key = policies.get("s3_secret_key")
            s3_requester_pays = policies.get("s3_requester_pays")
            s3_multi_keys = policies.get("s3_multi_keys")

            if not storage_gateway.s3_allow_multi_keys:
                if s3_multi_keys is not None and len(s3_multi_keys) > 0:
                    raise AttributeError(
                        "The storage gateway does not support multiple S3 keys."
                    )

            if storage_gateway.s3_user_credential_required:
                if s3_key_id:
                    display_name = s3_key_id
                else:
                    display_name = self.unconfigured_display_name
                if not policies:
                    policies = {
                        "DATA_TYPE": self.POLICY_DATA_TYPE,
                    }
                if not (s3_key_id and s3_secret_key):
                    invalid = True
                elif (
                    getattr(storage_gateway, "s3_requester_pays", False)
                    and not s3_requester_pays
                ):
                    raise AttributeError(
                        "The s3_requester_pays property is required for this credential."
                    )
                else:
                    provisioned = True
            else:
                display_name = self.anonymous_display_name
                if s3_key_id or s3_secret_key:
                    raise ValueError("key not allowed for anonymous access")
                if s3_requester_pays:
                    raise ValueError(
                        "s3_equester_pays not allowed for anonymous access"
                    )
                invalid = False
                provisioned = True
                policies = {
                    "DATA_TYPE": self.POLICY_DATA_TYPE,
                }
            doc = {
                **doc,
                **{
                    "display_name": display_name,
                    "invalid": invalid,
                    "deleted": False,
                    "provisioned": provisioned,
                    "policies": policies,
                },
            }
        super().__init__(
            appsync_data=appsync_data, doc=doc, storage_gateway=storage_gateway
        )

    @property
    def connector_id(self):
        return self.CONNECTOR_ID

    @connector_id.setter
    def connector_id(self, value):
        pass

    @property
    def s3_key_id(self) -> str:
        value = self.policies.get("s3_key_id")
        if value is None:
            raise AttributeError(
                "The s3_key_id property is not defined for this user credential"
            )
        return value

    @s3_key_id.setter
    def s3_key_id(self, value: t.Optional[str]) -> None:
        if self.storage_gateway.s3_user_credential_required:
            self.policies["s3_key_id"] = value
            if value is None:
                self.invalid = True
                self.display_name = "Unprovisioned S3 Credential"

    @s3_key_id.deleter
    def s3_key_id(self) -> None:
        self.policies["s3_key_id"] = None
        if self.storage_gateway.s3_user_credential_required:
            self.invalid = True
            self.display_name = "Unprovisioned S3 Credential"

    @property
    def s3_multi_keys(self):
        if not self.storage_gateway.s3_allow_multi_keys:
            return []

        # If the s3cred policy got updated to disallow do not include them in the list
        value = self.policies.get("s3_multi_keys")
        if value is None:
            return []
        return value

    @s3_multi_keys.setter
    def s3_multi_keys(self, value) -> None:
        if value is None:
            self.policies["s3_multi_keys"] = None
            return
        self.policies["s3_multi_keys"] = value

    @s3_multi_keys.deleter
    def s3_multi_keys(self) -> None:
        self.policies["s3_multi_keys"] = None

    @checks_policy_property(prop="s3_multi_keys")
    def check_s3_multi_keys(self, value) -> None:
        mandatory_fields = ["s3_key_id", "s3_secret_key", "path_prefixes"]
        s3_cred_map = self._make_cred_map()

        # initialize count to number of keys in the credential
        count = len(s3_cred_map)

        for p in value:
            # id always must be in the incoming document
            if "s3_key_id" not in p:
                raise AttributeError(
                    "Every s3_multi_keys entry requires a s3_key_id field"
                )
            if "s3_secret_key" in p and p["s3_secret_key"] is None:
                # update is deleting a key
                if p["s3_key_id"] in s3_cred_map:
                    count = count - 1
            elif p["s3_key_id"] not in s3_cred_map:
                # update is adding a key
                count = count + 1

            # check that all fields are either in the incoming document or already set
            key_id = p["s3_key_id"]
            for f in mandatory_fields:
                if f not in p:
                    try:
                        current = s3_cred_map[key_id]
                        _ = current[f]
                    except KeyError:
                        raise AttributeError(
                            f"s3_multi_keys entry requires the {f} field"
                        )

        if count > schema_s3.MAX_S3_CREDENTIAL_LIST:
            raise AttributeError(
                "The s3_multi_keys property may contain at most "
                f"{schema_s3.MAX_S3_CREDENTIAL_LIST} entries"
            )

    @checks_policy_property(prop="s3_key_id")
    def check_s3_key_id(self, value):
        if not self.storage_gateway.s3_user_credential_required and value:
            raise AttributeError()

    @property
    def s3_secret_key(self) -> str:
        value = self.policies.get("s3_secret_key")
        if value is None:
            raise AttributeError(
                "The s3_secret_key property is not defined for this user credential"
            )
        return value

    @s3_secret_key.setter
    def s3_secret_key(self, value):
        if self.storage_gateway.s3_user_credential_required:
            self.policies["s3_secret_key"] = value
            if value is None:
                self.invalid = True

    @s3_secret_key.deleter
    def s3_secret_key(self):
        self.policies["s3_secret_key"] = None
        if self.storage_gateway.s3_user_credential_required:
            self.invalid = True

    @checks_policy_property(prop="s3_secret_key")
    def check_s3_secret_key(self, value: str):
        if not self.storage_gateway.s3_user_credential_required and value:
            raise AttributeError()

    @property
    def s3_requester_pays(self) -> bool:
        value = self.policies.get("s3_requester_pays")
        if value is None:
            raise AttributeError(
                "The s3_requester_pays property is not defined for this user credential"
            )
        return value

    @s3_requester_pays.setter
    def s3_requester_pays(self, value):
        if self.storage_gateway.s3_user_credential_required:
            self.policies["s3_requester_pays"] = value
            if getattr(self.storage_gateway, "s3_requester_pays", False) and not value:
                self.invalid = True

    @s3_requester_pays.deleter
    def s3_requester_pays(self):
        self.policies["s3_requester_pays"] = None
        if getattr(self.storage_gateway, "s3_requester_pays", False):
            self.invalid = True

    @checks_policy_property(prop="s3_requester_pays")
    def check_s3_requester_pays(self, value: bool):
        if not self.storage_gateway.s3_user_credential_required and value:
            raise AttributeError()
        if getattr(self.storage_gateway, "s3_requester_pays", False) and not value:
            raise AttributeError(
                "The s3_requester_pays property is required for this credential."
            )

    def patch(self, **doc):
        doc.pop("display_name", None)
        policies = doc.get("policies", {})

        if self.storage_gateway.s3_user_credential_required:
            if policies is None:
                s3_key_id = None
                s3_secret_key = None
                s3_requester_pays = None
            else:
                s3_key_id = policies.get("s3_key_id", getattr(self, "s3_key_id", None))
                s3_secret_key = policies.get(
                    "s3_secret_key", getattr(self, "s3_secret_key", None)
                )
                s3_requester_pays = policies.get(
                    "s3_requester_pays", getattr(self, "s3_requester_pays", None)
                )

            if not s3_key_id or not s3_secret_key:
                if not doc.get("invalid", True):
                    raise ValueError(
                        "Cannot mark this credential as valid without keys"
                    )
                doc["invalid"] = True
            elif (
                getattr(self.storage_gateway, "s3_requester_pays", False)
                and not s3_requester_pays
            ):
                if not doc.get("invalid", True):
                    raise ValueError(
                        "Cannot mark this credential as valid without requester pays"
                    )
                doc["invalid"] = True
            else:
                doc["provisioned"] = True

            doc["display_name"] = self._provisioned_display_name(s3_key_id)

        super().patch(**doc)

    def _make_cred_map(self):
        s3_cred_map = {}
        for c in self.s3_multi_keys:
            s3_cred_map[c["s3_key_id"]] = c
        return s3_cred_map

    def _patch_s3_multi_keys(self, patch_s3_creds):
        s3_cred_map = self._make_cred_map()

        for s in patch_s3_creds:
            try:
                e_id = str(s["s3_key_id"])
                if "s3_secret_key" in s and s["s3_secret_key"] is None:
                    try:
                        del s3_cred_map[e_id]
                    except KeyError:
                        raise AttributeError(f"The id {e_id} was not found")
                else:
                    if e_id not in s3_cred_map:
                        s3_cred_map[e_id] = {}
                    for k in s.keys():
                        s3_cred_map[e_id][k] = s[k]
            except KeyError:
                raise ValueError("Each s3_multi_keys patch entry must have an id")

        self.s3_multi_keys = s3_cred_map.values()

    def _patch_policies(self, **policies):
        DATA_TYPE = policies.get("DATA_TYPE", self.POLICY_DATA_TYPE)
        if DATA_TYPE not in self.POLICY_DATA_TYPES:
            raise ValueError("Invalid DATA_TYPE {}".format(DATA_TYPE))
        _connector_properties = self.CONNECTOR_POLICIES.get(
            DATA_TYPE, self.connector_properties
        )
        _validate_s3_credential_in_policy(policies, patch=True)
        for p in policies:
            if p in _connector_properties:
                if p == "s3_multi_keys":
                    self._patch_s3_multi_keys(policies["s3_multi_keys"])
                else:
                    setattr(self, p, policies[p])
            elif p != "DATA_TYPE":
                raise ValueError("Unknown policy {}".format(p))

    def put(self, **doc):
        invalid = doc.get("invalid")
        policies = doc.get("policies", {})
        s3_key_id = policies.get("s3_key_id")
        s3_secret_key = policies.get("s3_secret_key")
        s3_requester_pays = policies.get("s3_requester_pays")
        _validate_s3_credential_in_policy(policies, patch=False)

        if self.storage_gateway.s3_user_credential_required:
            if not (s3_key_id and s3_secret_key):
                if invalid is False:
                    raise ValueError("Cannot mark this credential as valid")
                invalid = True
            elif (
                getattr(self.storage_gateway, "s3_requester_pays", False)
                and not s3_requester_pays
            ):
                raise AttributeError(
                    "The s3_requester_pays property is required for this credential."
                )
            else:
                self.provisioned = True
        else:
            if s3_key_id or s3_secret_key:
                raise ValueError("key not allowed for anonymous access")
            if s3_requester_pays:
                raise ValueError("requester_pays not allowed for anonymous access")
            self.provisioned = True
        if invalid is None:
            invalid = False

        doc["invalid"] = invalid
        doc["display_name"] = self._provisioned_display_name(s3_key_id)

        super().put(**doc)

    def _provisioned_display_name(self, s3_key_id):
        if self.storage_gateway.s3_user_credential_required:
            if s3_key_id:
                display_name = s3_key_id
            else:
                display_name = self.unconfigured_display_name
        else:
            display_name = self.anonymous_display_name

        return display_name

    def delete(self):
        self.s3_secret_key = None
        self.s3_requester_pays = None
        self.s3_multi_keys = None
        super().delete()

    def __repr__(self):
        return "<S3UserCredential {}>".format(self.id)
