from __future__ import annotations

import asyncio
import enum
import json
import os
import socket
import typing as t
from abc import ABC, abstractmethod
from base64 import b64encode
from threading import Thread
from urllib.parse import urlparse
from uuid import uuid4

from flask import Flask
from typing_extensions import Literal
from websockets import (
    connect as ws_connect,
    exceptions as ws_exceptions,
)
from websockets.asyncio.client import ClientConnection as WebSocketClientConnection

if t.TYPE_CHECKING:
    import collections.abc as abc

    from websockets import Subprotocol

    DecodedMessageHandler = abc.Callable[[dict[str, t.Any]], None]
    EncodedMessageHandler = abc.Callable[[t.Union[str, bytes]], None]


def _setup_proxy_connection(
    appsync_url: str,  # wss:// to appsync
) -> socket.socket | None:
    """
    Setup a proxy connection to appsync if HTTPS_PROXY or https_proxy are set in
    the environment.

    If proxying is required, returns socket.socket that has been configured for
    tunneling, otherwise returns None.

    Note: We allow exceptions to be raised here and caught by our caller.
    """

    #
    # Prefer HTTPS_PROXY over https_proxy (for no particular reason)
    #
    proxy_url = os.getenv("HTTPS_PROXY")
    if proxy_url is None:
        proxy_url = os.getenv("https_proxy")
    if proxy_url is None:
        return None  # No proxy config found, skipping

    #
    # Break proxy_url out into scheme and netloc then validate.
    #
    parsed_proxy_url = urlparse(proxy_url)
    if parsed_proxy_url.netloc == "":
        raise ValueError(f"Invalid proxy URL: {proxy_url}")
    if parsed_proxy_url.scheme != "http":
        raise ValueError(
            "Appsync proxy support only for HTTP proxies, "
            f"{parsed_proxy_url.scheme} requested",
        )

    #
    # Breakout the proxy IP/Host and port
    #

    # "127.0.0.1" or "127.0.0.1:8080" or "localhost" or "localhost:8080"
    ip_host_port = parsed_proxy_url.netloc.split(":")
    # IP/Host is always available at index 0
    proxy_ip_host = ip_host_port[0]
    # Port may not be available, default to 80
    proxy_port = int(ip_host_port[1]) if len(ip_host_port) == 2 else 80

    #
    # Now let's prepare the proxy for tunneling
    #
    sock = socket.create_connection((proxy_ip_host, proxy_port))

    parsed_appsync_url = urlparse(appsync_url)

    proxy_request = (
        f"CONNECT {parsed_appsync_url.netloc}:443 HTTP/1.1\r\n"
        f"Host: {parsed_appsync_url.netloc}:443\r\n\r\n"
    )
    sock.send(str.encode(proxy_request))
    # We expect a response like b'HTTP/1.1 200 Connection established\r\n\r\n'
    response = sock.recv(4096)
    if not response.startswith(b"HTTP/1.1 200"):
        sock.close()
        raise ValueError(f"Unexpected proxy response: {response!r}")

    return sock


class SubscriptionBase(ABC):
    """
    Abstract subscription base class designed to allow alternate websocket
    implementations to be dropped in as replacements.
    """

    # Seconds we'll wait past the keepalive timeout before disconnecting
    __TIMEOUT_GRACE_SECS = 30

    #
    # Public interface.
    #

    def __init__(
        self,
        info: dict[str, str],
        access_token: str,
        endpoint_id: str,
        current_app_obj: Flask,
        on_message: DecodedMessageHandler | None = None,
        on_disconnect: abc.Callable[[], None] | None = None,
    ) -> None:
        """
        Initialize the subscription (websocket) connection to appsync's realtime interface. After
        this call to initialize the Subscription instance, the caller will start to receive data
        on the on_message and on_disconnect callbacks. The caller can terminate the subscription at
        any time by calling disconnect().

        Note: when an error is received, the error message handler will log the error and terminate
        the event loop which results in a on_disconnect() callback.

        info: dict returned from 'GET /subscribe' which contains the WSS URL and HOST informaton
        for the subscription connection.

        access_token: auth bearer token used for the API gateway and appsync authorizers

        endpoint_id: the ID of the endpoint to subscribe to notifications for

        current_app_obj: the object behind Flask's current_app proxy for the calling thread, used
        for logging and setting the current_app context for callbacks

        on_message: called when the subscription returns updates for the endpoint from appsync's
        realtime interface

        on_disconnect: called when the websocket has fully terminated
        """
        self._info = info
        self._access_token = access_token
        self._endpoint_id = endpoint_id
        self._current_app_obj = current_app_obj
        self._on_message_callback = on_message
        self._on_disconnect_callback = on_disconnect
        self._connection_timeout_secs = 0
        # This is unique per SubscriptionBase instance.
        self._sub_id = str(uuid4())

    def disconnect(self) -> None:
        """
        Terminate the websocket. This will trigger the on_disconnect() callback. It is not
        necessary to call this function if the on_disconnect() callback has already triggered.
        It is safe to call this from an external thread (the internal thead can
        use _stop_event_loop()).
        """
        self._disconnect()

    #
    # Private helpers.
    #
    def __header_encode(self, header: t.Any) -> str:
        return b64encode(json.dumps(header).encode("utf-8")).decode("utf-8")

    def __appsync_url(self) -> str:
        api_header = {
            "host": self._info["HOST"],
            "Authorization": self._access_token,
        }
        return (
            self._info["WSS_URL"]
            + "?header="
            + self.__header_encode(api_header)
            + "&payload=e30="
        )

    def _subscribe_msg(self, endpoint_id: str) -> str:
        gql_subscription = {
            "query": (
                "subscription ($owner_id: ID!) { "
                "Endpoint(owner_id: $owner_id) { "
                "owner_id deleted configuration_objects { "
                "items { "
                "id "
                "tag "
                "serial "
                "created_at "
                "updated_at "
                "configuration "
                "deleted "
                "} "
                "next_page_token }"
                "}"
                "}"
            ),
            "variables": {"owner_id": endpoint_id},
        }

        register = {
            "id": self._sub_id,
            "payload": {
                "data": json.dumps(gql_subscription),
                "extensions": {
                    "authorization": {
                        "host": self._info["HOST"],
                        "Authorization": self._access_token,
                    }
                },
            },
            "type": "start",
        }
        start_sub = json.dumps(register)
        return start_sub

    #
    # Lifecycle events for the subscription.
    #

    def _start(self) -> None:
        """
        Called from the subclass to initiate the event loop in a detached
        thread.
        """
        Thread(target=self.__thread_entry, daemon=True).start()

    def __thread_entry(self) -> None:
        """
        Private method to launch the subclass's event loop and handle the
        on_disconnect message when the loop ends.
        """
        # This call will not return until the websocket is dead.
        self._run_event_loop(
            # mypy wants this cast but if the type isn't quoted, the tests will
            # hang because the thread dies on
            #     "NameError: name 'Subprotocol' is not defined"
            self.__appsync_url(),
            [t.cast("Subprotocol", "graphql-ws")],
        )

        # Notify the caller that we have disconnected
        if self._on_disconnect_callback:
            with self._current_app_obj.app_context():
                self._on_disconnect_callback()

    def _on_open(self) -> None:
        """
        Called from the subclass when the websocket first opens.
        """
        # Send the connection init message
        init = {"type": "connection_init"}
        init_conn = json.dumps(init)
        self._send_message(init_conn)

    def _on_message(self, encoded_msg: bytes | str) -> None:
        """
        Process messages sent to us from subclass websocket implmentations.
        The encoded_msg looks like:
         {
             'type': <see msg_handlers keys>,
             ... (msg specific data)
         }
        """

        class MsgType(enum.Enum):
            # The msg should be passed encoded as a string
            ENCODED = enum.auto()
            # The msg should be passed decoded as a dict
            DECODED = enum.auto()

        MessageHandler = t.Union[
            tuple[Literal[MsgType.ENCODED], "EncodedMessageHandler"],
            tuple[Literal[MsgType.DECODED], "DecodedMessageHandler"],
        ]
        msg_handlers: dict[str, MessageHandler] = {
            "connection_ack": (MsgType.DECODED, self.__on_connection_ack_message),
            "start_ack": (MsgType.DECODED, self.__on_start_ack_message),
            "data": (MsgType.DECODED, self.__on_data_message),
            "ka": (MsgType.DECODED, self.__on_keepalive_message),
            "connection_error": (MsgType.ENCODED, self.__on_error_message),
            "error": (MsgType.ENCODED, self.__on_error_message),
        }
        default_handler: MessageHandler = (MsgType.ENCODED, self._on_unknown_message)

        handler = default_handler
        decoded_msg: dict[str, t.Any] = {}

        try:
            # msgs that can't be decoded get the default handler
            decoded_msg = json.loads(encoded_msg)
            # msgs that aren't known get the default handler
            msg_type = decoded_msg.get("type")
            if msg_type:
                handler = msg_handlers[msg_type]
            else:
                handler = default_handler
        except Exception:
            pass

        if handler[0] == MsgType.DECODED:
            handler[1](decoded_msg)
        else:
            handler[1](encoded_msg)

    def __keepalive_timer_callback(self) -> None:
        """
        Called whenever a connection keepalive timer expires. The websocket
        connection is dead, we need to shut it down.
        """
        self._stop_event_loop()

    #
    # Abstract method declarations.
    #

    @abstractmethod
    def _run_event_loop(
        self, appsync_url: str, subprotocols: t.Sequence[Subprotocol] | None
    ) -> None:
        """
        Launch the websocket. This method should not return until the websocket
        is dead.
        """
        pass

    @abstractmethod
    def _disconnect(self) -> None:
        """
        Called by an external thread to shutdown the event loop in a thread-safe way.
        """
        pass

    @abstractmethod
    def _stop_event_loop(self) -> None:
        """
        Shutdown the websocket and exit the event loop.
        """
        pass

    @abstractmethod
    def _send_message(self, msg: str) -> None:
        """
        Send a message over the websocket.
        """
        pass

    @abstractmethod
    def _reset_keepalive_timer(
        self, connection_timeout_secs: int, timer_callback: abc.Callable[[], None]
    ) -> None:
        """
        Cancel any outstanding keepalive timer and register a new one.
        """
        pass

    #
    # Generic message handling. Private.
    #

    def __on_connection_ack_message(self, msg: dict[str, t.Any]) -> None:
        """
        This message is received in response to our connection_init request
        when the websocket first opened. The msg will look like this:

         {
             'type': 'connection_ack',
             'payload': {
                 'connectionTimeoutMs': 300000
             }
         }
        """
        # Record the connection timeout, we use it to configure keepalive
        connection_timeout_ms = msg["payload"]["connectionTimeoutMs"]
        self._connection_timeout_secs = connection_timeout_ms / 1000

        # Register our subscription
        self._send_message(self._subscribe_msg(self._endpoint_id))

        # Set a keepalive timer
        self._reset_keepalive_timer(
            # +__TIMEOUT_GRACE_SECS seconds to avoid any edge cases
            self._connection_timeout_secs + SubscriptionBase.__TIMEOUT_GRACE_SECS,
            self.__keepalive_timer_callback,
        )

    def __on_start_ack_message(self, msg: dict[str, t.Any]) -> None:
        """
        This message is sent in response to our subscribe request which we sent
        on connection_ack. The msg will look like this:

         {
             'id': '40d7d955-8dd1-4ee0-b203-8fdb4cbc697b',
             'type': 'start_ack'
         }

        'id' should be equal to self._sub_id which we sent when we subscribed.
        """
        # We do not need to respond to this message.
        return

    def __on_data_message(self, msg: dict[str, t.Any]) -> None:
        """
        These messages are the updates we subscribed to for our endpoint. They
        look like this:
         {
           'id': '<unique id>,
           'type': 'data',
           'payload': {'data': {'Endpoint': {...}}}
         }
        """
        if self._on_message_callback:
            with self._current_app_obj.app_context():
                self._on_message_callback(msg["payload"])

    def __on_keepalive_message(self, msg: dict[str, t.Any]) -> None:
        """
        These messages are sent to signal that the websocket connection is
        still stable. We need to reset the timer based on
        self._connection_timeout_secs. The keepalive messages look like:
         {
           'type': 'ka'
         }
        """
        self._reset_keepalive_timer(
            # +__TIMEOUT_GRACE_SECS seconds to avoid any edge cases
            self._connection_timeout_secs + SubscriptionBase.__TIMEOUT_GRACE_SECS,
            self.__keepalive_timer_callback,
        )

    def __on_error_message(self, msg: bytes | str) -> None:
        """
        This is called whenever we receive an error mesasge over the websocket.
        For example:
         {
           'id': '5da1452b-4cb5-4a4f-8e38-84f48e103c84',
           'type': 'error',
           'payload': {
             'errors': [
               {
                 'errorType': 'Unauthorized',
                 'message': 'Not Authorized to access Endpoint on type Subscription'
               }
             ]
           }
         }
        This also receives errors about the connection generated by the websocket
        implementation.
        """
        # We can't rely on the msg format since we accept errors from appsync and
        # the websocket implementation.
        self._current_app_obj.logger.error(f"Appsync error message. {msg!r}")
        self._stop_event_loop()

    def _on_unknown_message(self, msg: bytes | str) -> None:
        """
        The websocket sent us a message type that we were not expecting. This is usually
        a usage error or an internal exception.
        """
        self._current_app_obj.logger.error(f"Appsync unexpected message: {msg!r}")
        self._stop_event_loop()


class SubscriptionWebsockets(SubscriptionBase):
    """
    websocket subscription client based on the websockets package.
    """

    def __init__(
        self,
        info: dict[str, str],
        access_token: str,
        endpoint_id: str,
        current_app_obj: Flask,
        on_message: DecodedMessageHandler | None = None,
        on_disconnect: abc.Callable[[], None] | None = None,
    ) -> None:
        self._websocket: WebSocketClientConnection | None = None
        self._event_loop: asyncio.AbstractEventLoop | None = None
        self._keepalive_timer: asyncio.TimerHandle | None = None

        super().__init__(
            info, access_token, endpoint_id, current_app_obj, on_message, on_disconnect
        )
        super()._start()

    #
    # Implementations of super's abstractmethods.
    #

    def _run_event_loop(
        self, appsync_url: str, subprotocols: t.Sequence[Subprotocol] | None
    ) -> None:
        """
        Launch the websocket. This method should not return until the websocket
        is dead.
        """
        self._event_loop = asyncio.new_event_loop()
        coroutine = self.__process_event_loop(appsync_url, subprotocols)
        self._event_loop.create_task(coroutine)
        # run_forever() will exit when someone calls self._event_loop.stop()
        self._event_loop.run_forever()
        # Now we need to close down the websocket otherwise asyncio tasks will run
        # in the background until the process exits.
        self._event_loop.run_until_complete(self.__close())
        # This will terminate the asyncio event_loop itself and free resources
        self._event_loop.close()

    def _disconnect(self):
        """
        Called by an external thread to shutdown the event loop in a thread-safe way.
        """
        if self._event_loop and not self._event_loop.is_closed():
            self._event_loop.call_soon_threadsafe(self._stop_event_loop)

    def _stop_event_loop(self) -> None:
        """
        Shutdown the websocket and exit the event loop.
        """
        if self._event_loop and not self._event_loop.is_closed():
            # This will kick us out of the asyncio loop so that run_forever() will return
            self._event_loop.stop()

    def _send_message(self, msg: str) -> None:
        """
        Queue a send task on our event loop.
        """
        assert self._event_loop is not None
        self._event_loop.create_task(self.__send(msg))

    def _reset_keepalive_timer(
        self, connection_timeout_secs: int, timer_callback: abc.Callable[[], None]
    ) -> None:
        """
        Cancel any outstanding keepalive timer and register a new one.
        """
        # Cancel the old timer
        if self._keepalive_timer:
            self._keepalive_timer.cancel()

        # Register the new timer
        assert self._event_loop is not None

        self._keepalive_timer = self._event_loop.call_later(
            connection_timeout_secs, timer_callback
        )

    #
    # Private helpers.
    #
    async def __process_event_loop(
        self,
        appsync_url: str,
        subprotocols: t.Sequence[Subprotocol] | None,
    ) -> None:
        self._websocket = await self.__connect(appsync_url, subprotocols)
        if self._websocket is None:
            self._stop_event_loop()
            return

        # After connect, launch our recv() handler
        asyncio.ensure_future(self.__recv_loop())

        # Inform super that the websocket is now open
        self._on_open()

    async def __recv_loop(self) -> None:
        while self._websocket is not None:
            msg = None
            try:
                msg = await self.__recv()
            except ws_exceptions.ConnectionClosed:
                return

            # Forward onto super's on_message() method.
            if msg:
                self._on_message(msg)

    async def __connect(
        self,
        appsync_url: str,  # wss:// to appsync
        subprotocols: t.Sequence[Subprotocol] | None,
    ) -> WebSocketClientConnection | None:
        try:
            sock = _setup_proxy_connection(appsync_url)
            return await ws_connect(appsync_url, subprotocols=subprotocols, sock=sock)
        except Exception as e:
            self._current_app_obj.logger.debug(
                f"Websocket failed to connect: [{type(e)}] {e}"
            )
            return None

    async def __send(self, msg, *args, **kwargs) -> None:
        if self._event_loop and not self._event_loop.is_closed():
            self._current_app_obj.logger.debug(f"AppSync message sent: {msg}")
            assert self._websocket is not None
            return await self._websocket.send(msg, *args, **kwargs)

    async def __recv(self, *args, **kwargs) -> bytes | str | None:
        if self._event_loop and not self._event_loop.is_closed():
            assert self._websocket is not None
            msg = await self._websocket.recv()
            self._current_app_obj.logger.debug(f"AppSync message received: {msg!r}")
            return msg
        return None

    async def __close(self, *args, **kwargs) -> None:
        if self._websocket is not None:
            return await self._websocket.close()


# Determines which websocket implementation to use
Subscription = SubscriptionWebsockets
