from __future__ import annotations

import datetime
import typing as t
from itertools import count, takewhile

import globus.manager.api.models as models
from globus.manager.api.blueprints.utils import delete_object
from globus.manager.constants import ConfigurationObjectTag, NodeStatus

if t.TYPE_CHECKING:
    import sched

    import flask

    from .exclusive_subscription_lock import ExclusiveSubscriptionLock


class AutoDeleteManager:
    """
    Delete unused guest collections.
    """

    # Number of deletes that are processed per poll
    AUTO_DELETE_LIMIT = 5
    # Number of minutes to wait before retying if this node is not active
    AUTO_DELETE_INACTIVE_DELAY_MINS = 5
    # Number of minutes to wait before retying if auto delete encounters an error
    AUTO_DELETE_ERROR_DELAY_MINS = 5
    # Number of minutes to wait before retying if there are remaining guest
    # collections to delete and we are just pausing to release the DB lock
    AUTO_DELETE_PAUSE_DELAY_MINS = 1
    # Number of minutes to wait before retying if there are no remaining guest
    # collections to delete
    AUTO_DELETE_FINISHED_DELAY_MINS = 60

    def __init__(
        self,
        scheduler: sched.scheduler,
        app: flask.Flask,
        subscription_lock: ExclusiveSubscriptionLock,
    ):
        """
        Initialize an AutoDeleteManager object to automatic deletion of unused
        guest collections.

        :param sched.scheduler scheduler:
            A scheduler which this class will use to register timed events.
        :param GCSManagerApp app:
            An app which this class will use to interact with appsync, the
            database, locks and logs.
        :param: subscription_lock
            Lock to use when processing query results. This locks out the GCS
            manager process from processing requests that would interact with
            the database while we're processing changes that might make many
            changes to the database.
        """
        self.scheduler = scheduler
        "Scheduler to register events with"

        self.app = app
        "Flask application this object is a part of"

        self.__subscription_lock = subscription_lock
        "Intra+Inter process subscription (db) lock"

        self.scheduler.enter(0, 0, self.auto_delete_guests)

    def _get_active_node_position(self) -> tuple[int, int]:
        """
        Returns an tuple of (index, count):

        index: this active node's position in an ordered list of all active nodes
        in the endpoint.

        count: total number of currently active nodes in the endpoint

        Note: This should only be called from an active node.
        """

        active_nodes_ids: list[str] = [
            node.id
            for node in models.Endpoint.query.one().globus_dns_domain.nodes
            if node.status == NodeStatus.active
        ]
        self.app.logger.debug(
            f"multi-node endpoint active node ID list: {active_nodes_ids}"
        )

        try:
            active_node_index = active_nodes_ids.index(self.app.node.id)
            active_node_count = len(active_nodes_ids)
        except ValueError:
            # This should never happen because of the previous node status check
            # but ... future proofing
            self.app.logger.warning(
                "Auto collection delete failure: our active node id is not "
                "in the active nodes ids list"
            )
            raise

        return active_node_index, active_node_count

    def _should_guest_be_deleted(
        self,
        guest: models.Collection,
    ) -> bool:
        """
        Helper that returns True if guest is a candidate for auto deletion.

        Note: guest is assumed to be a guest collection (duh)
        """
        self.app.logger.debug(f"auto_delete_guests examining collection {guest.id}")
        if guest.collection_type != "guest":
            self.app.logger.debug(
                "The collection checker for auto delete was called with "
                "mapped collection"
            )
            return False

        # Grab auto_delete_timeout, if set
        timeout = getattr(guest.mapped_collection, "auto_delete_timeout", None)

        # Skip this guest collection auto_delete_timeout was not set
        if timeout is None:
            self.app.logger.debug(
                f"collection {guest.id} is not configured for auto delete"
            )
            return False

        # Skip this guest collection if configured to do so
        if getattr(guest, "skip_auto_delete", False) is True:
            self.app.logger.debug(
                f"collection {guest.id} is configured to skip auto delete"
            )
            return False

        # If last_update is not set, use created_at
        last_access: datetime.date = (
            guest.last_access
            if guest.last_access is not None
            else guest.created_at_date
        )
        now = datetime.datetime.now(datetime.timezone.utc).date()
        should_be_deleted: bool = ((now - last_access)).days >= timeout
        self.app.logger.debug(
            f"collection {guest.id} is "
            f"{'not ' if should_be_deleted is False else ''}ready for auto delete"
        )

        return should_be_deleted

    def _lookup_guests_to_delete(
        self,
        limit: int,
        selector: t.Callable[[models.Collection], bool] | None = None,
    ) -> list[models.Collection]:
        """
        Return a list of collections, up to a count of limit, for which selector
        returns True.
        """
        query = models.Collection.query.filter(
            ~models.Collection.is_deleted,  # type: ignore
            models.Collection.collection_type == "guest",
        )

        # This will create/return a list of guest collections filtered by the optional
        # selector, filtered by _should_get_be_deleted() returning True, and limit the
        # list length to limit.
        c = count()
        return list(
            takewhile(
                lambda x: next(c) < limit,
                (
                    guest
                    for guest in query.all()
                    if (
                        (selector(guest) if selector else True)
                        and self._should_guest_be_deleted(guest)
                    )
                ),
            )
        )

    def _multi_node_guest_selector(self) -> t.Callable[[models.Collection], bool]:
        # This will tell us which node we are in an ordered list of active
        # nodes.
        node_index, node_count = self._get_active_node_position()

        # In case an assistant is down even though the node has an active
        # status, we'll apply an offset to node_index based on hour of the
        # day. In this way, we can pick up a down assistant's guest
        # deletions in the next hour or so.
        hour_of_day = datetime.datetime.now().hour

        def _active_node_guest_selector(guest: models.Collection) -> bool:
            return (
                (hash(guest.id) + node_index + hour_of_day) % node_count
            ) == node_index

        return _active_node_guest_selector

    def _delete_guest_collection(
        self: AutoDeleteManager,
        guest: models.Collection,
    ) -> None:
        self.app.logger.info(
            f"Auto deleting guest collection {guest.id}. "
            f"Last access: {guest.last_access} "
            f"Auto delete timeout: {guest.auto_delete_timeout} "
            f"Skip auto delete: {getattr(guest, 'skip_auto_delete', False)}"
        )

        ep_serial = (
            models.db.session.query(models.ConfigurationObject.serial)
            .filter(models.ConfigurationObject.tag == ConfigurationObjectTag.ENDPOINT)
            .scalar()
        )

        delete_object(guest, update=True)
        collection_serial = guest.serial

        for r in models.Role.query.filter(models.Role.resource == guest.id):
            self.app.logger.debug(f"Creating PendingUpdate to delete role {r}")
            pending_update = models.PendingUpdate(
                object_id=r.id,
                change_time=datetime.datetime.now(tz=datetime.timezone.utc),
                endpoint_serial=ep_serial,
                collection_serial=collection_serial,
                update_type=models.PendingUpdate.DELETE,
            )
            models.db.session.add(pending_update)
        pending_update = models.PendingUpdate(
            object_id=guest.id,
            change_time=datetime.datetime.now(tz=datetime.timezone.utc),
            endpoint_serial=ep_serial,
            collection_serial=collection_serial,
            update_type=models.PendingUpdate.DELETE,
        )
        models.db.session.add(pending_update)
        models.db.session.commit()

    def auto_delete_guests(self) -> None:
        # Smaller wrapper to keep us from repeating this try...except throughout
        # the class
        try:
            self._auto_delete_guests()
        except Exception as e:
            self.app.logger.warning(
                f"Auto collection delete failure {e}, rescheduling auto delete in "
                f"{self.AUTO_DELETE_ERROR_DELAY_MINS} minutes"
            )
            self.scheduler.enter(
                self.AUTO_DELETE_ERROR_DELAY_MINS * 60,
                0,
                self.auto_delete_guests,
            )

    def _auto_delete_guests(self) -> None:
        """
        Event to delete unused guest collections.
        """
        # Future: Should we care about pending updates?
        with (
            self.app.app_context(),
            self.__subscription_lock.exclusive(),
            models.db.session.no_autoflush,
        ):
            self.app.logger.debug("auto_delete_guests enter")

            # If this node is not active, just reschedule. Although we could likely
            # process auto deletes successfully in a single-node endpoint, I think
            # it's best to give the admin the impression that an inactive node is
            # truly inactive and not doing unexpected work.
            self.app.logger.debug("auto_delete_guests checking that node is active")
            if self.app.node.status != NodeStatus.active:
                self.app.logger.debug(
                    "node is not active, rescheduling collection auto delete in "
                    f"{self.AUTO_DELETE_INACTIVE_DELAY_MINS} minutes"
                )
                self.scheduler.enter(
                    self.AUTO_DELETE_INACTIVE_DELAY_MINS * 60,
                    0,
                    self.auto_delete_guests,
                )
                return

            # In a multi node endpoint, we want to try to avoid duplicating efforts
            # of other assistants. This is what the selector is used for; it will
            # apply a filter that limits this assistant's actions to a unique set
            # of guest collections.

            self.app.logger.debug("auto_delete_guests choosing selector")
            # selector=None removes all per-node assistant filtering
            selector: t.Callable[[models.Collection], bool] | None = None
            if len(models.Endpoint.query.one().globus_dns_domain.nodes) > 1:
                selector = self._multi_node_guest_selector()

            self.app.logger.debug(
                "auto_delete_guests using the "
                + ("single " if selector is None else "multi ")
                + "node selector"
            )

            # Add 1 to the limit so that we know if we need to wake up sooner rather
            # than later to process additional guest collections
            self.app.logger.debug("auto_delete_guests searching for guests to delete")
            guests_to_delete = self._lookup_guests_to_delete(
                AutoDeleteManager.AUTO_DELETE_LIMIT + 1,
                selector,
            )

            # Delete each unused guest collection
            for guest in guests_to_delete[: AutoDeleteManager.AUTO_DELETE_LIMIT]:
                # soft delete collections and roles, etc
                self._delete_guest_collection(guest)

            # Base the amount of time to sleep before retry on whether there are
            # remaining guests this assistant should delete.
            msg = "has completed, retrying in"
            delay = self.AUTO_DELETE_FINISHED_DELAY_MINS
            if len(guests_to_delete) > self.AUTO_DELETE_LIMIT:
                msg = "pausing for"
                delay = self.AUTO_DELETE_PAUSE_DELAY_MINS

            self.app.logger.debug(f"Auto collection delete {msg} {delay} minutes")
            self.scheduler.enter(delay * 60, 0, self.auto_delete_guests)
