"""
globus-connect-server node enable

The `globus-connect-server node enable` command sets the current node's status
to active and adds it to the Globus DNS wildcard domain. The machine must have
been previously setup with `node setup`.
"""

from __future__ import annotations

import json
import os

import click
from sqlalchemy.exc import OperationalError

import globus.manager.api.errors as errors
import globus.manager.api.models as m
from globus.manager.api.app import make_app
from globus.manager.api.blueprints.utils import update_object
from globus.manager.scripts.utils.setup_utils import load_keychain
from globus.manager.utils.keychain import NodeKey


def node_enable(
    ctx: click.Context,
    verbose: bool,
) -> None:
    if verbose:
        os.environ["GCS_MANAGER_LOG_LEVEL"] = "DEBUG"

    if os.getuid() != 0:
        click.echo("This command must be run as the root user", err=True)
        ctx.exit(1)

    app = make_app(app_config="globus.manager.api.config.node_enable")
    m.db.init_app(app)

    with app.app_context():
        try:
            for c in m.Configuration.query.all():
                if c.name == "node_key":
                    app.config[c.name.upper()] = NodeKey(**json.loads(c.value))
                else:
                    app.config[c.name.upper()] = c.value
        except OperationalError:
            raise click.ClickException(
                "This command must be run on a configured GCSv5 node"
            )

        for name in [
            "CLIENT_ID",
            "CLIENT_SECRET",
            "ENVIRONMENT",
            "NODE_KEY",
            "NODE_ID",
        ]:
            if name not in app.config:
                raise click.ClickException(
                    "This command must be run on a configured GCSv5 node"
                )

        # Load the keychain from appsync
        load_keychain()

        with app.appsync_client.pause_notifications():
            wildcard_domain = m.GlobusDnsDomain.query.one_or_none()
            if not wildcard_domain:
                raise errors.InternalServerError("Server is misconfigured")

        patch = {
            "DATA_TYPE": "node#1.2.0",
            "id": app.config["NODE_ID"],
            "status": "active",
        }
        update_object(wildcard_domain, lambda o, d: o.patch_node(d), patch)
        click.echo(f"Message: Updated node {app.config['NODE_ID']}")
