"""
globus-connect-server node new-secret

The globus-connect-server node new-secret command updates the Auth client secret
associated with this node, deleting the previous node secret if it was a unique
one.
"""
from __future__ import annotations

import os
import textwrap

import click
from sqlalchemy.orm.exc import NoResultFound

import globus_sdk

from globus.manager.api.app import make_app
from globus.manager.api.models import Configuration
from globus.manager.database import db


def node_new_secret(
    agree_to_delete_previous_secret: bool,
) -> None:
    if os.getuid() != 0:
        raise click.ClickException("This command must be run as the root user")

    app = make_app(app_config="globus.manager.api.config.node_cli")
    db.init_app(app)

    with app.app_context():
        try:
            client_id = Configuration.query.filter_by(name="client_id").one().value
            client_secret_config = Configuration.query.filter_by(
                name="client_secret"
            ).one()
            client_secret = client_secret_config.value
            node_id = Configuration.query.filter_by(name="node_id").one().value
            node_secret_id_config = Configuration.query.filter_by(
                name="node_secret_id"
            ).one_or_none()
            globus_environment = (
                Configuration.query.filter_by(name="environment").one().value.lower()
            )
        except NoResultFound:
            raise click.ClickException(
                "Globus Connect Server is not configured. Run "
                '"globus-connect-server node setup" to enable the services on this node.'
            )
        app.config["ENVIRONMENT"] = globus_environment
        os.environ["GLOBUS_SDK_ENVIRONMENT"] = globus_environment
        old_node_secret_id: str | None = None
        if node_secret_id_config:
            old_node_secret_id = node_secret_id_config.value

            # Warn the admin that the node's previous node-specific key will be
            # deleted and that it will interrupt GCS services until the manager
            # and the assistant are restarted.
            if not agree_to_delete_previous_secret and not click.confirm(
                "\n".join(
                    textwrap.wrap(
                        "This node's current secret will be deleted. The GCS services "
                        "on this node will not be operational until the GCS manager "
                        "and assistant are restarted. You can skip this prompt with "
                        "the --agree-to-delete-previous-secret option. Do you want to "
                        "continue?"
                    )
                ),
                default=False,
                show_default=True,
            ):
                click.echo("Aborting...")
                click.get_current_context().exit(1)

        try:
            auth_token = globus_sdk.ClientCredentialsAuthorizer(
                globus_sdk.ConfidentialAppAuthClient(
                    client_id,
                    client_secret,
                ),
                ["urn:globus:auth:scope:auth.globus.org:manage_projects"],
            ).access_token
        except globus_sdk.AuthAPIError as e:
            if e.code != "UNAUTHORIZED":
                raise

            click.echo(
                "\n".join(
                    [""]
                    + textwrap.wrap(
                        "Error authenticating to Globus Auth with the node's "
                        "current credentials. This can occur if client_id or "
                        "secret have been deleted."
                    )
                    + [""]
                    + textwrap.wrap(
                        f"Verify client_id {client_id} and secret ID"
                        f"{' ' + old_node_secret_id if old_node_secret_id else ''} "
                        "on {}".format(
                            globus_sdk.config.get_service_url(
                                "auth",
                                environment=os.environ.get(
                                    "GLOBUS_SDK_ENVIRONMENT", "production"
                                ),
                            )
                        )
                    )
                    + [""]
                    + textwrap.wrap(
                        "If the client_id and default secret still exist, you "
                        "can use 'globus-connect-server node setup' to recover."
                    )
                ),
                err=True,
            )
            click.get_current_context().exit(1)

        if not auth_token:
            raise click.ClickException("Unable to obtain access token.")

        client = globus_sdk.AuthClient(
            authorizer=globus_sdk.AccessTokenAuthorizer(auth_token),
            app_name="globus-connect-server node new-secret",
        )
        credential_doc = {"credential": {"name": f"Node credential for {node_id}."}}

        try:
            response = client.post(
                f"/v2/api/clients/{client_id}/credentials", data=credential_doc
            )
        except Exception as e:
            raise click.ClickException(f"FOO: {e}")
        except globus_sdk.GlobusAPIError as e:
            raise click.ClickException(f"Failed to create client credentials: {e}")
        credential_secret = response["credential"]["secret"]
        credential_id = response["credential"]["id"]
        client_secret_config.value = credential_secret
        db.session.add(client_secret_config)
        if not node_secret_id_config:
            node_secret_id_config = Configuration(
                name="node_secret_id", value=credential_id
            )
        else:
            node_secret_id_config.value = credential_id
        db.session.add(node_secret_id_config)
        db.session.commit()

        if old_node_secret_id:
            try:
                client.delete(
                    f"/v2/api/clients/{client_id}/credentials/{old_node_secret_id}"
                )
            except globus_sdk.GlobusAPIError as e:
                raise click.ClickException(
                    f"Failed to delete old secret {old_node_secret_id}: {e}"
                )

            click.echo(
                "\n".join(
                    [""]
                    + textwrap.wrap(
                        "The GCS manager and assistant must be restarted. If "
                        "those services are currently running, they will be "
                        "non-functional until restarted."
                    )
                )
            )
        else:
            click.echo(
                "\n".join(
                    [""]
                    + textwrap.wrap(
                        "The GCS manager and assistant must be restarted in order "
                        "for the new secret to take affect. Until the services are "
                        "restarted, they will continue to use the previous secret."
                    )
                )
            )
