from __future__ import annotations

import logging
import time
import typing as t
from collections import namedtuple

import josepy
from acme import client, crypto_util, messages
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from OpenSSL import crypto

logger = logging.getLogger(__name__)

AcmeCertificate = namedtuple(
    "AcmeCertificate", ["domain_name", "certificate", "chain", "key"]
)


class AcmeClient:
    # TODO: more verbose user agent? full version? client id?
    USER_AGENT: t.ClassVar[str] = "GlobusConnectServer"
    ACC_KEY_BITS: t.ClassVar[int] = 2048
    CERT_PKEY_BITS: t.ClassVar[int] = 2048
    WAIT_SEC: t.ClassVar[int] = 120

    ACME_DIRECTORY_URL: t.ClassVar[
        str
    ] = "https://acme-v02.api.letsencrypt.org/directory"

    def __init__(
        self,
        dns_client,
        acme_key: dict | None = None,
        acme_account: dict | None = None,
    ) -> None:
        """
        Create an AcmeClient that can be used for certificate request and renewal.

        If acme_key is not passed in, it is auto generated.
        If acme_account is not passed in, then it is generated from
        acme.client.ClientV2(directory, net=net).new_account().
        """
        self.dns_client = dns_client
        self._set_acme_key(acme_key)
        self._set_acme_account(acme_account)

    @property
    def acme_key(self) -> dict:
        """
        Used for serializing, reinstantiating and testing.
        """
        return self._acme_key.to_json()

    def _set_acme_key(self, key: dict | None) -> None:
        """
        One-time initialization of the acme key.
        """
        # We will only set it once, regardless of value
        if hasattr(self, "_acme_key"):
            raise AttributeError("can't set attribute")

        if key:
            # Cast this because mypy sees from_json as returning
            # TypedJSONObjectWithFields although it's called as a subclass
            # of JWKRSA.
            self._acme_key = t.cast(josepy.JWKRSA, josepy.JWKRSA.from_json(key))
        else:
            self._acme_key = josepy.JWKRSA(
                key=rsa.generate_private_key(
                    public_exponent=65537,
                    key_size=self.ACC_KEY_BITS,
                    backend=default_backend(),
                )
            )

    @property
    def acme_account(self) -> dict:
        """
        Used for serializing, reinstantiating and testing.
        """
        return self._acme_account.to_json()

    def _set_acme_account(self, account: dict | None) -> None:
        """
        One-time initialization of the acme account.
        """
        # We will only set it once, regardless of value
        if hasattr(self, "_acme_account"):
            raise AttributeError("can't set attribute")

        if not account:
            self._acme_account = self._acme_client.new_account(
                messages.NewRegistration.from_data(terms_of_service_agreed=True)
            )
        else:
            self._acme_account = messages.RegistrationResource.from_json(account)

    @property
    def _acme_client(self) -> client.ClientV2:
        """
        Private member function to create an underlying ACME client for internal use.
        """
        if getattr(self, "__acme_client", None) is None:
            net = client.ClientNetwork(
                self._acme_key,
                account=getattr(self, "acme_account", None),
                user_agent=self.USER_AGENT,
            )
            directory = messages.Directory.from_json(
                net.get(self.ACME_DIRECTORY_URL).json()
            )
            self.__acme_client = client.ClientV2(directory, net=net)
        return self.__acme_client

    def get_cert_chain_and_key(self, domain_name: str) -> AcmeCertificate:
        """
        Looks up gcs_certificate, certificate_chain, and certificate_key from
        config. If gcs_certificate needs renewal or doesn't exist yet requests a
        new cert from letsencrypt
        """
        # create private key for the certificate
        # we use a new private key
        pkey = crypto.PKey()
        pkey.generate_key(crypto.TYPE_RSA, self.CERT_PKEY_BITS)
        key = crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey)

        # create Certificate Signing Request order
        csr = crypto_util.make_csr(key, [domain_name, "*." + domain_name])
        order = self._acme_client.new_order(csr)

        # get DNS challenges
        dns_challenges = []
        for authz in order.authorizations:
            for i in authz.body.challenges:
                # Find the supported challenge.
                if i.chall.typ == "dns-01":
                    dns_challenges.append(i)
        if len(dns_challenges) != 2:
            raise ValueError("Expected 2 DNS challenges")

        # get validation values to place in TXT records
        txt = []
        for chal in dns_challenges:
            txt.append(chal.validation(self._acme_client.net.key))

        # update TXT records
        logger.info("Creating TXT records with values: {}".format(txt))
        self.dns_client.create_txt(domain_name, txt)

        # wait for DNS propagation
        logger.info(
            "Waiting {} seconds for DNS changes " "to propagate".format(self.WAIT_SEC)
        )
        time.sleep(self.WAIT_SEC)

        # let letsencrypt know we are ready for validation
        for chal in dns_challenges:
            response = chal.response(self._acme_client.net.key)
            self._acme_client.answer_challenge(chal, response)

        # poll and finalize order to get fullchain
        logger.info("Polling and finalizing order")
        finalized_order = self._acme_client.poll_and_finalize(order)

        # split full chain into cert and chain
        fullchain_pem = finalized_order.fullchain_pem
        cert = crypto.dump_certificate(
            crypto.FILETYPE_PEM,
            # mypy expects fullchain_pem to be bytes not str
            crypto.load_certificate(
                crypto.FILETYPE_PEM, fullchain_pem.encode(encoding="utf-8")
            ),
        ).decode()
        chain = fullchain_pem[len(cert) :].lstrip()

        decoded_key = key.decode()

        return AcmeCertificate(domain_name, cert, chain, decoded_key)
